/* This software is part of an Instructables project called
 * Wordclock - No Special Tools Required! Easy to Build
 * It's up to you to use, modify and share it.
 * Please refer to Instructables for a detailled guidance
 * Attribution Non-commercial Share Alike (by-nc-sa)
 * 
 * Developped by
 * Wolfgang Berger (aka Eman@home aka Bad_Wolf_Laboratories)
 * Besigheim, Germany 2021
 * Date           Version Change
 * March,11 2021  6.2     Bugfix webinterface NTP time display
 * March,17 2021  7.0     Bugfix Hour display german standard, minute leds
 */
#include <ESP8266WiFi.h>
#include <WiFiUdp.h>
#include <WiFiClient.h>
#include <NTPClient.h>
#include <ESP8266HTTPClient.h>
#include <ESP8266WebServer.h>
#include <EEPROM.h>
#include <TimeLib.h>
#include <NeoPixelBus.h>

//WIFI and Web
//////////////
ESP8266WebServer  server(80);
WiFiClient  client;
WiFiUDP     ntpUDP;

String      NTP_Server = "europe.pool.ntp.org"; // or "pool.ntp.org";

NTPClient   timeClient(ntpUDP, NTP_Server.c_str(), 3600, 60000);

//display time
//////////////
String      NTPtime;
String      DST_mode; //auto / summer / winter
int         summer_offset, winter_offset;
bool        summer;
String      language;

//Network
/////////
String  ssid1, pass1, ssid2, pass2;
String  Accesspoint_Name; 
bool    AP_running;
bool wifiConnected = false;
bool reconnection_allowed = false; //set to true, if there was a connection possible

IPAddress Accesspoint_IP(192, 168, 4, 1);
IPAddress Accesspoint_netmask(255, 255, 255, 0);
  
//Neopixels
///////////
#define colorSaturation 128
const uint16_t PixelCount = 100;
const uint8_t PixelPin = 2;  // ignored for Esp8266, but necessary for function call
//connect ESP GPI02 pin D1 mini = D4

NeoPixelBus<NeoGrbFeature, NeoEsp8266Uart1800KbpsMethod> strip(PixelCount, PixelPin); //Wemos D1 mini

RgbColor red(128, 0, 0);
RgbColor orange(127,50,0);
RgbColor brown(50,25,7);
RgbColor green(0, 128, 0);
RgbColor blue(0, 0, 128);
RgbColor white(128);
RgbColor purple(100,0,127); 
RgbColor black(0);

RgbColor this_color(black);

String color = "";  //friendly color name

//system
////////
unsigned long last_second = 99; //99 to force startup
unsigned long last_minute = 99; //99 to force startup
unsigned long last_hour   = 99; //99 to sync

int j;

//********************************************************

void setup() 
{
  EEPROM.begin(512);
  Serial.begin(115200);
  delay(500);         //wait for serial startup
  timeClient.begin();
  StartupNeopixels();

  GetConfiguration(); 
  
  WiFi.mode(WIFI_AP_STA);   
  StartupAccesspoint(); 
  connectWifi();
  
  ConfigWebServer();
  
  if (wifiConnected) 
  {
    CalculateDST(); //first
    GetNtpTime();   //second
    ShowMyIP();     //show IP on clock display
  }
  Serial.println("Startup done");
}

//********************************************************

void loop() 
{
  once_per_second();  //all the things we want to do once per second
  once_per_minute();  //all the things we want to do once per minute
  once_per_hour();    //all the things we want to do once per hour
}

//********************************************************

void once_per_second()
{
  if (second() != last_second)
  {
    last_second = second();

    if (!wifiConnected) 
    {
      strip.SetPixelColor(99, red);
      strip.Show(); 
      
      if(reconnection_allowed) connectWifi(); //connection lost, try reconnection
    }
    
    server.handleClient();                    //check for browser request and handle it
     
    shutdown_accesspoint(5 * 60000);         //shutdown accesspoint 5 minutes after power on   
  }
}

void once_per_minute()    //will be forced by HTML post
{
  if (minute() != last_minute)
  {
    last_minute = minute();

    GetConfiguration();   //refresh configuration
    
    if (wifiConnected) 
    {
      CalculateDST(); //call this first
      GetNtpTime();   //call this second
    }
    
    if(reconnection_allowed)  //we are or we were connected: show time
    {
      AllLedsOff();           //prepare display: all leds off
      TimeToText();           //                 set text LEDs 
      MinuteLeds();           //                 set minute LEDs
    }
    strip.Show();             // display it
  }
}

void once_per_hour() //nothing to do in this version
{
  if (hour() != last_hour)
  {
    last_hour = hour();
    //your code, what you want to do once per hour
  }
}

/* 
LED layout
first row left to right, 
second row right to left, and so on
De              En
 0    5          0
 ES IST          IT'SZA
15        6     15        6
 FÜNFZEHN--      TWENTYFIVE
16        25    16        25
 NACHBIS---      QUARTERTEN
35        26    35        26
 DREIHALB--      HALF PASTO
36        45    36        45
 VIERTEL---      ONETWOFOUR
55        46    55        46
 EINSASECHS      THREEFIVET
56        65    56        65
 ZWEICZEHNE      SIXELEVENE
75        66    75        66
NEUNHDREIL       SEVENINE-N
76        85    76        85
FÜNFTZWÖLF       EIGHTWELVE
95        86    95        86
SIEBENVIER       O'CLOCK---
96  99          96  99
 oooo            oooo
*/

void TimeToText()
{
  Serial.println("Time to display:");
  Serial.print(hour());Serial.print(":");Serial.print(minute());Serial.print(":");Serial.print(second());
  if(language == "de")
  {
    for(j = 0 ; j <= 5 ; j++) strip.SetPixelColor(j, this_color); // ES IST
    strip.SetPixelColor(2, black);
    
    Minutes_Text_de( minute() - (minute() % 5)); //minutes 5, 10, 15, ...
    Hours_Text_de(hour());
  }  
  if(language == "de1")
  {
    for(j = 0 ; j <= 5 ; j++) strip.SetPixelColor(j, this_color); // ES IST
    strip.SetPixelColor(2, black);
    
    Minutes_Text_de1( minute() - (minute() % 5)); //minutes 5, 10, 15, ...
    Hours_Text_de(hour());
  }
  if(language == "en")
  {
    for(j = 0 ; j <= 3 ; j++) strip.SetPixelColor(j, this_color); // IT'S
    Minutes_Text_en( minute() - (minute() % 5)); //minutes 5, 10, 15, ...
    Hours_Text_en(hour());
  }
}

void Hours_Text_de(int hours)
{
  if(language == "de") //German standard
    if(minute() > 19) hours = hours + 1;  //8:10 zehn nach 8, 8:15 fünfzehn nach 8
  if(language == "de1") //German South
    if(minute() > 14) hours = hours + 1; //8:10 zehn nach 8, 8:15 viertel neun
  
  if (hours > 12) hours = hours - 12;
  if (hours == 0) hours = 12;
  show_number_de(hours);
}

void Hours_Text_en(int hours)
{
  if(language == "en") //English
    if(minute() > 34) hours = hours + 1; //8:30 half past 8, 8:35 twentyfive to 9  
  
  if (hours > 12) hours = hours - 12;
  if (hours == 0) hours = 12;
  show_number_en(hours);
}

void show_number_de(int number)
{  
  switch (number)
  {
    case  0:
      strip.SetPixelColor(75, this_color);  
      strip.SetPixelColor(73, this_color);  
      strip.SetPixelColor(66, this_color);  
      strip.SetPixelColor(84, this_color);      
      break;
    case  1:
      for(j = 52 ; j <= 55 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    case  2:
      for(j = 56 ; j <= 59 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    case  3:
      for(j = 67 ; j <= 70 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    case  4:
      for(j = 86 ; j <= 89 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    case  5:
      for(j = 76 ; j <= 79 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    case  6:
      for(j = 46 ; j <= 50 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    case  7:
      for(j = 90 ; j <= 95 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    case  8:
      strip.SetPixelColor(51, this_color);  
      strip.SetPixelColor(60, this_color);  
      strip.SetPixelColor(71, this_color);  
      strip.SetPixelColor(80, this_color);      
      break;
    case  9:
      for(j = 72 ; j <= 75 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    case 10:
      for(j = 61 ; j <= 64 ; j++) strip.SetPixelColor(j, this_color);   
      break;
    case 11:
      strip.SetPixelColor(65, this_color);  
      strip.SetPixelColor(66, this_color);  
      strip.SetPixelColor(85, this_color);      
      break;
    case 12:
      for(j = 81 ; j <= 85 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    default:
      break;
  }  
}

void Minutes_Text_de(int minutes)
{
  switch (minutes)
  {
    case  5: 
      for(j = 12 ; j <= 15 ; j++) strip.SetPixelColor(j, this_color); //fünf
      for(j = 16 ; j <= 19 ; j++) strip.SetPixelColor(j, this_color); //nach
      break;
    case 10: 
      for(j = 8 ; j <= 11 ; j++) strip.SetPixelColor(j, this_color);  //zehn
      for(j = 16 ; j <= 19 ; j++) strip.SetPixelColor(j, this_color); //nach
      break;
    case 15: 
      for(j = 8 ; j <= 15 ; j++) strip.SetPixelColor(j, this_color); //fünfzehn
      for(j = 16 ; j <= 19 ; j++) strip.SetPixelColor(j, this_color); //nach
      break;
    case 20: 
      for(j = 8 ; j <= 11 ; j++) strip.SetPixelColor(j, this_color);  //zehn
      for(j = 20 ; j <= 22 ; j++) strip.SetPixelColor(j, this_color); //bis
      for(j = 28 ; j <= 31 ; j++) strip.SetPixelColor(j, this_color); //halb
      break;
    case 25: 
      for(j = 12 ; j <= 15 ; j++) strip.SetPixelColor(j, this_color); //fünf
      for(j = 20 ; j <= 22 ; j++) strip.SetPixelColor(j, this_color); //bis
      for(j = 28 ; j <= 31 ; j++) strip.SetPixelColor(j, this_color); //halb
      break;
    case 30: 
      for(j = 28 ; j <= 31 ; j++) strip.SetPixelColor(j, this_color); //halb
      break;
    case 35: 
      for(j = 12 ; j <= 15 ; j++) strip.SetPixelColor(j, this_color); //fünf
      for(j = 16 ; j <= 19 ; j++) strip.SetPixelColor(j, this_color); //nach
      for(j = 28 ; j <= 31 ; j++) strip.SetPixelColor(j, this_color); //halb
      break;
    case 40: 
      for(j = 8 ; j <= 11 ; j++) strip.SetPixelColor(j, this_color);  //zehn
      for(j = 16 ; j <= 19 ; j++) strip.SetPixelColor(j, this_color); //nach
      for(j = 28 ; j <= 31 ; j++) strip.SetPixelColor(j, this_color); //halb
      break;
    case 45: 
      for(j = 8 ; j <= 15 ; j++) strip.SetPixelColor(j, this_color); //fünfzehn
      for(j = 20 ; j <= 22 ; j++) strip.SetPixelColor(j, this_color); //bis
      break;
    case 50: 
      for(j = 8 ; j <= 11 ; j++) strip.SetPixelColor(j, this_color);  //zehn
      for(j = 20 ; j <= 22 ; j++) strip.SetPixelColor(j, this_color); //bis
      break;
    case 55: 
      for(j = 12 ; j <= 15 ; j++) strip.SetPixelColor(j, this_color); //fünf
      for(j = 20 ; j <= 22 ; j++) strip.SetPixelColor(j, this_color); //bis
      break;
    default: 
      break;
  }  
}

void Minutes_Text_de1(int minutes)
{
  switch (minutes)
  {
    case  5: 
      for(j = 12 ; j <= 15 ; j++) strip.SetPixelColor(j, this_color); //fünf
      for(j = 16 ; j <= 19 ; j++) strip.SetPixelColor(j, this_color); //nach
      break;
    case 10: 
      for(j = 8 ; j <= 11 ; j++) strip.SetPixelColor(j, this_color);  //zehn
      for(j = 16 ; j <= 19 ; j++) strip.SetPixelColor(j, this_color); //nach
      break;
    case 15: 
      for(j = 36 ; j <= 42 ; j++) strip.SetPixelColor(j, this_color); //viertel
      break;
    case 20: 
      for(j = 12 ; j <= 15 ; j++) strip.SetPixelColor(j, this_color); //fünf
      for(j = 16 ; j <= 19 ; j++) strip.SetPixelColor(j, this_color); //nach
      for(j = 36 ; j <= 42 ; j++) strip.SetPixelColor(j, this_color); //viertel
      break;
    case 25: 
      for(j = 12 ; j <= 15 ; j++) strip.SetPixelColor(j, this_color); //fünf
      for(j = 20 ; j <= 22 ; j++) strip.SetPixelColor(j, this_color); //bis
      for(j = 28 ; j <= 31 ; j++) strip.SetPixelColor(j, this_color); //halb
      break;
    case 30: 
      for(j = 28 ; j <= 31 ; j++) strip.SetPixelColor(j, this_color); //halb
      break;
    case 35: 
      for(j = 12 ; j <= 15 ; j++) strip.SetPixelColor(j, this_color); //fünf
      for(j = 16 ; j <= 19 ; j++) strip.SetPixelColor(j, this_color); //nach
      for(j = 28 ; j <= 31 ; j++) strip.SetPixelColor(j, this_color); //halb
      break;
    case 40: 
      for(j = 12 ; j <= 15 ; j++) strip.SetPixelColor(j, this_color); //fünf
      for(j = 20 ; j <= 22 ; j++) strip.SetPixelColor(j, this_color); //bis
      for(j = 32 ; j <= 42 ; j++) strip.SetPixelColor(j, this_color); //dreiviertel
      break;
    case 45: 
      for(j = 32 ; j <= 42 ; j++) strip.SetPixelColor(j, this_color); //dreiviertel
      break;
    case 50: 
      for(j = 12 ; j <= 15 ; j++) strip.SetPixelColor(j, this_color); //fünf
      for(j = 16 ; j <= 19 ; j++) strip.SetPixelColor(j, this_color); //nach
      for(j = 32 ; j <= 42 ; j++) strip.SetPixelColor(j, this_color); //dreiviertel
      break;
    case 55: 
      for(j = 12 ; j <= 15 ; j++) strip.SetPixelColor(j, this_color); //fünf
      for(j = 20 ; j <= 22 ; j++) strip.SetPixelColor(j, this_color); //bis
      break;
    default: 
      break;
  }  
}


void show_number_en(int number)
{  
  switch (number)
  {
    case  0: //ZERO
      strip.SetPixelColor( 4, this_color);  
      strip.SetPixelColor(21, this_color);  
      strip.SetPixelColor(22, this_color);  
      strip.SetPixelColor(26, this_color);      
      break;
    case  1:
      for(j = 36 ; j <= 38 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    case  2:
      for(j = 39 ; j <= 41 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    case  3:
      for(j = 51 ; j <= 55 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    case  4:
      for(j = 42 ; j <= 45 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    case  5:
      for(j = 47 ; j <= 50 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    case  6:
      for(j = 56 ; j <= 58 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    case  7:
      for(j = 71 ; j <= 75 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    case  8:
      for(j = 76 ; j <= 80 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    case  9:
      for(j = 68 ; j <= 71 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    case 10:
      strip.SetPixelColor(46, this_color);  
      strip.SetPixelColor(65, this_color);  
      strip.SetPixelColor(66, this_color);
      break;
    case 11:
      for(j = 59 ; j <= 64 ; j++) strip.SetPixelColor(j, this_color);         
      break;
    case 12:
      for(j = 80 ; j <= 85 ; j++) strip.SetPixelColor(j, this_color);    
      break;
    default:
      break;
  }  
}

void Minutes_Text_en(int minutes)
{
  switch (minutes)
  {
    case  0: 
      for(j = 89 ; j <= 95 ; j++) strip.SetPixelColor(j, this_color); //o'clock
      break;
    case  5: 
      for(j =  6 ; j <=  9 ; j++) strip.SetPixelColor(j, this_color); //five
      for(j = 27 ; j <= 30 ; j++) strip.SetPixelColor(j, this_color); //past
      break;
    case 10: 
      for(j = 23 ; j <= 25 ; j++) strip.SetPixelColor(j, this_color); //ten
      for(j = 27 ; j <= 30 ; j++) strip.SetPixelColor(j, this_color); //past

      break;
    case 15: 
      strip.SetPixelColor(5, this_color);                             //a
      for(j = 16 ; j <= 22 ; j++) strip.SetPixelColor(j, this_color); //quarter
      for(j = 27 ; j <= 30 ; j++) strip.SetPixelColor(j, this_color); //past
      break;
    case 20: 
      for(j = 10 ; j <= 15 ; j++) strip.SetPixelColor(j, this_color); //twenty
      for(j = 27 ; j <= 30 ; j++) strip.SetPixelColor(j, this_color); //past
      break;
    case 25: 
      for(j =  6 ; j <= 15 ; j++) strip.SetPixelColor(j, this_color); //twentyfive
      for(j = 27 ; j <= 30 ; j++) strip.SetPixelColor(j, this_color); //past
      break;
    case 30: 
      for(j = 32 ; j <= 35 ; j++) strip.SetPixelColor(j, this_color); //half
      for(j = 27 ; j <= 30 ; j++) strip.SetPixelColor(j, this_color); //past
      break;
    case 35: 
      for(j =  6 ; j <= 15 ; j++) strip.SetPixelColor(j, this_color); //twentyfive
      for(j = 26 ; j <= 27 ; j++) strip.SetPixelColor(j, this_color); //to
      break;
    case 40: 
      for(j = 10 ; j <= 15 ; j++) strip.SetPixelColor(j, this_color); //twenty
      for(j = 26 ; j <= 27 ; j++) strip.SetPixelColor(j, this_color); //to
      break;
    case 45: 
      strip.SetPixelColor(5, this_color);                             //a
      for(j = 16 ; j <= 22 ; j++) strip.SetPixelColor(j, this_color); //quarter
      for(j = 26 ; j <= 27 ; j++) strip.SetPixelColor(j, this_color); //to
      break;
    case 50: 
      for(j = 23 ; j <= 25 ; j++) strip.SetPixelColor(j, this_color); //ten
      for(j = 26 ; j <= 27 ; j++) strip.SetPixelColor(j, this_color); //to
      break;
    case 55: 
      for(j =  6 ; j <=  9 ; j++) strip.SetPixelColor(j, this_color); //five
      for(j = 26 ; j <= 27 ; j++) strip.SetPixelColor(j, this_color); //to
      break;
    default: 
      break;
  }  
}

void ShowMyIP()
{
  IPAddress myIP = WiFi.localIP();
  String ThisIP = String(myIP[0],DEC) + "." + String(myIP[1],DEC) + "." + String(myIP[2],DEC) + "." + String(myIP[3],DEC);
  String digit;

  Serial.print("IP address: ");
  Serial.println(ThisIP);
  
  for (int j = 0 ; j < ThisIP.length() ; j++) 
  {
    AllLedsOff();
    digit = ThisIP.charAt(j);

    if( ThisIP.charAt(j) == '.')
    {
      AllLedsOff();
      strip.Show();
      delay(500);
    }
    else
    {
   if(language == "en") //English
      show_number_en (digit.toInt());
   else                 // German
    show_number_de (digit.toInt());
    }
    
    strip.Show();
    delay(700);
  }
  
  AllLedsOff();
  strip.Show();
  delay(500);
}

void MinuteLeds()
{
  int MinuteLEDs = minute() % 5;
  if(MinuteLEDs >= 1)   strip.SetPixelColor(96, this_color);  
  if(MinuteLEDs >= 2)   strip.SetPixelColor(97, this_color);  
  if(MinuteLEDs >= 3)   strip.SetPixelColor(98, this_color);  
  if(MinuteLEDs == 4)   strip.SetPixelColor(99, this_color); 
}

void AllLedsOff()
{
  for (j=0; j < PixelCount ;j++) strip.SetPixelColor(j, black);
}

void connectWifi()
{
  if (!wifiConnected) connectWLAN(ssid1, pass1); 
  if (!wifiConnected) connectWLAN(ssid2, pass2); 
  if (!wifiConnected) Serial.println("not connected to WIFI");
  if (wifiConnected) reconnection_allowed = true;
}

void connectWLAN(String ssid, String pass) 
{
  bool connection = false;

  Serial.println("");
  Serial.println("connect to: ");
  Serial.println(ssid);
  WiFi.begin(ssid.c_str(), pass.c_str());

  for (j=0 ; j<20 ; j++)
  {
    if (WiFi.status() == WL_CONNECTED)
    {
      Serial.println("");
      Serial.println("connected");
      Serial.println(WiFi.localIP());
      connection = true;
      break;
    }
    strip.SetPixelColor(99, red);
    strip.Show();
    delay(100);
    strip.SetPixelColor(99, black);
    strip.Show();
    delay(500);
  }
  wifiConnected = connection;
}
/*
eeprom layout
 00...   9  identifier, contains text "eeprom_ok" after first run
 10...  29  ssid1
 30...  59  pass1
 60...  89  ssid2
 90...  119  pass2
120...  124  language
125...  134  color
135...  144  dst mode
145...  148  summer offset
149...  152  winter offset
155...  174  accesspoint name
*/

void GetConfiguration()
{
  if(ee_read_string(  0) != "config_ok")  //a new ESP does not have the magic word in eeprom
  {
    Serial.println("reset Config");       //so we preset it to valid values
    ee_write_string(  0, "config_ok");

    ee_write_string( 10, "wlan_1_ssid");  //will be overwritten by web config
    ee_write_string( 30, "wlan_1_pwd");
    ee_write_string( 60, "wlan_2_ssid");  //second Wifi credentials for workshop area
    ee_write_string( 90, "wlan_2_pwd");   //no access from outside
    ee_write_string(120, "de");           //language for display de or en
    ee_write_string(125, "red");          //color 
    ee_write_string(135, "auto");         //DST: summer, winter, auto (central Europe)
    ee_write_string(145, "1");            //summer time offset
    ee_write_string(149, "0");            //winter time offset
    ee_write_string(155, "word_clock");   //accesspoint name
  }
  
  ssid1             = ee_read_string( 10);
  pass1             = ee_read_string( 30);
  ssid2             = ee_read_string( 60);
  pass2             = ee_read_string( 90);
  language          = ee_read_string(120);
  color             = ee_read_string(125);
  
  DST_mode          = ee_read_string(135);
  summer_offset     = ee_read_string(145).toInt();
  winter_offset     = ee_read_string(149).toInt();
  Accesspoint_Name  = ee_read_string(155);

  
  if (color == "green")   this_color = green;
  if (color == "red")     this_color = red;
  if (color == "orange")  this_color = orange;
  if (color == "brown")   this_color = brown;
  if (color == "blue")    this_color = blue;
  if (color == "white")   this_color = white;
  if (color == "purple")  this_color = purple;

  if( Accesspoint_Name == "") Accesspoint_Name = "wordclock"; // in case of config error
}

String ee_read_string(int adr)
{
  const int MAXREAD = 30;
  char c;
  String ee_string = "";
    
  for (int j = 0; j < MAXREAD; j++)
  { 
    c = char(EEPROM.read(adr + j));
    if(c == '\0') break;
    ee_string += c; 
  } 
  return(ee_string);
}

void ee_write_string(int adr, String value)
{
  int i;
  for (i = 0; i < (value.length() +1); ++i)
    {
      EEPROM.write(adr +i, value[i]);
    }
  EEPROM.write(adr +i , '\0');
  EEPROM.commit();
}

void StartupAccesspoint()
{
  WiFi.softAPConfig(Accesspoint_IP, Accesspoint_IP, Accesspoint_netmask);
  WiFi.softAP(Accesspoint_Name.c_str());
  
  AP_running = true;
  Serial.println("Accesspoint started");
}

void shutdown_accesspoint(unsigned long runtime)
{
  if (AP_running == true)
  {
    if(millis() >  runtime)
    {
      Serial.println("Accesspoint shutdown");
      WiFi.softAPdisconnect();
      AP_running = false;
    }
  }
}

void StartupNeopixels()
{
  strip.Begin();
  strip.Show();
  Serial.println("Neopixel bus started");
}

void CalculateDST() //Central European Summer Time (CEST)
{
  if(DST_mode == "auto")
  {
    summer = false;
    int thisYear, lastSundayInMarch, lastSundayInOctober;
    String now_Time, Summer_begin, Summer_end;
    
    timeClient.update();
    setTime( timeClient.getEpochTime());
    thisYear = year();
      
    now_Time = String(year(),DEC);
    if(month() < 10) now_Time = now_Time + "0";
    now_Time = now_Time + String(month(),DEC);
    if(day() < 10) now_Time = now_Time + "0";
    now_Time = now_Time + String(day(),DEC);
    if(hour() < 10) now_Time = now_Time + "0";
    now_Time = now_Time + String(hour(),DEC);
    if(minute() < 10) now_Time = now_Time + "0";
    now_Time = now_Time + String(minute(),DEC);
    
    setTime(02, 00, 00, 31, 03, thisYear);  // set to last day in March actual year
    lastSundayInMarch = 31 - weekday() +1;    //last sunday in march
    Summer_begin = String(year(),DEC) + "03" + String(lastSundayInMarch,DEC) + "0200";
    
    setTime(02, 00, 00, 31, 10, thisYear);  // set to last day in October actual year
    lastSundayInOctober = 31 - weekday() +1;      //last sunday in october
    Summer_end = String(year(),DEC) + "10" + String(lastSundayInOctober,DEC) + "0300";
    
    if(now_Time > Summer_begin) summer = true;
    if(now_Time > Summer_end) summer = false;
  }
  if(DST_mode == "summer") summer = true;
  if(DST_mode == "winter") summer = false;
  Serial.println("DST calculated");
}

void GetNtpTime()
{
  int j;

  if (wifiConnected)
  {
    for (j = 0; j < 5; j++)
    {
      timeClient.update();

      setTime( timeClient.getEpochTime());
      NTPtime = (String) day() + "." + (String) month() + "." + (String) year()  + " " +  (String) hour() + ":"; 
      if(minute() < 10) NTPtime += "0";
      NTPtime += (String) minute(); 
          
      if(summer)
      {
        setTime( timeClient.getEpochTime() + (summer_offset * SECS_PER_HOUR)); // summer
      }
      else
      {
        setTime( timeClient.getEpochTime() + (winter_offset * SECS_PER_HOUR)); // winter
      }
      if(year() > 2000) break;
      delay(1000);
    }
  
    if(year() < 2000)
      {
        Serial.println("NTP failure");
        last_minute = 99; // retry next minute
      }

    else
      Serial.println("NTP sync done");
      Serial.println(NTPtime);
  }
  else
    Serial.println("not connected, Time not correct");
}


void ConfigWebServer()
{
  server.on("/",              handleRoot);
  server.on("/factoryreset",  factoryreset); //call factory reset
  server.onNotFound(          handleNotFound);
  server.begin();
  Serial.println("HTTP server started");
}

void factoryreset()
{
  ee_write_string(  0, "resetted"); //force factory reset
  server.send(200, "text/html", restart_HTML());  
  delay(2000);
  ESP.restart();
}

void handleRoot() 
{  
  Serial.println("");

  if(server.client().localIP() == Accesspoint_IP) //we come via wordclock's access point
  {                                               //////////////////////////////////////
    Serial.println("* handle AP Root *");
    if(server.args() > 0) AP_Posted_Data();
    server.send(200, "text/html", AP_Root_HTML());
  }
    
  if(server.client().localIP() == WiFi.localIP()) //we come via wlan
  {                                               //////////////////
    Serial.println("* handle STA Root *");
    if(server.args() > 0) STA_Posted_Data();
    server.send(200, "text/html", STA_Root_HTML());
  }
}

void AP_Posted_Data()
{
  if(server.arg("ok") == "ok")
  {
    if( (server.arg("ssid1") != "") && (server.arg("pass1") != "")) 
    {
      ee_write_string( 10, server.arg("ssid1"));
      ee_write_string( 30, server.arg("pass1"));

      wifiConnected = false;                    //force connect to wifi

      Serial.println("reset AP and restart");
      WiFi.softAPdisconnect();
      AllLedsOff();
      server.send(200, "text/html", restart_HTML());
      ESP.restart(); //restart all, reset ESP after flashing, otherwise it will not work
    }
  }
}

void STA_Posted_Data()
{
  if(server.arg("ok") == "ok")
  {
    if(server.arg("language")       != "") ee_write_string(120, server.arg("language"));
    if(server.arg("color")          != "") ee_write_string(125, server.arg("color"));
    if(server.arg("DST_mode")       != "") ee_write_string(135, server.arg("DST_mode"));
    if(server.arg("summer_offset")  != "") ee_write_string(145, server.arg("summer_offset"));
    if(server.arg("winter_offset")  != "") ee_write_string(149, server.arg("winter_offset"));

    GetConfiguration(); //maybe we changed something
    last_minute = 99;   //force refresh wordclock's display
        
    if((server.arg("Accesspoint_Name") != Accesspoint_Name) && (server.arg("Accesspoint_Name") != ""))
    {
      Serial.println("reset AP and restart");
      WiFi.softAPdisconnect();
      ee_write_string(155, server.arg("Accesspoint_Name"));
      AllLedsOff();
      ESP.restart(); //restart all
    }
  }
}

void handleNotFound()
{
  String message = "Page Not Found\n\n";
  message += "URI: ";
  message += server.uri();
  server.send(404, "text/plain", message);  
}

String restart_HTML()
{
  String
  txt  = "<html><head><meta name=viewport content=""width=device-width, initial-scale=1""></head><body><font face=verdana>";
  txt += "Restart now</font></body></html>"; 
  return txt;
}

String AP_Root_HTML()
{
  String 
  txt  = "<html><head><meta name=viewport content=""width=device-width, initial-scale=1""></head>";
  txt += "<body><font face=verdana>";
  txt += "<form method=post>";
  txt += "select WLAN and enter password <BR>";
  txt += "wordclock will display the ip address after restart<BR>";

  byte numSsid = WiFi.scanNetworks();
  String this_ssid;
  for (int thisNet = 0; thisNet<numSsid; thisNet++) 
  {
    this_ssid = WiFi.SSID(thisNet);
    txt += "<br><input type=radio name=ssid1 value = " + this_ssid + ">" + this_ssid ;
  }

  txt += "<BR><BR>Password&nbsp;<input type=password name=pass1 value =" + pass1 + "><BR><BR>";
  txt += "<input type=submit name=ok value=ok>";
  txt += "</form></font></body></html>";

  return txt;
}

String STA_Root_HTML()
{
  String 
  txt  = "<html><head><meta name=viewport content=""width=device-width, initial-scale=1""></head>";
  txt += "<body><font face=verdana>";
  txt += "<form method=post>";

  txt += "<table>";
  txt += "<tr><td colspan = 2><font color=blue>Word Clock configuration</font></td></tr>";
  txt += "<tr><td>&nbsp;</td></tr>";
  
  txt += "<tr><td>NTP server</td><td>" + NTP_Server;
  txt += "</td></tr>";
  txt += "<tr><td>NTP Time</td><td>" + NTPtime + "</td></tr>";
  txt += "<tr><td>&nbsp;</td></tr>";
  
  txt += "<tr valign=top><td>Language (suitable text mask required!)</td>";
  txt += "<td>";
  txt += "<input type=radio name=language value = de  "; if(language == "de" ) txt += "checked"; txt += ">German standard<BR>";
  txt += "<input type=radio name=language value = de1 "; if(language == "de1") txt += "checked"; txt += ">German South<BR>";
  txt += "<input type=radio name=language value = en  "; if(language == "en" ) txt += "checked"; txt += ">English<BR>";
  txt += "</td></tr>";
  txt += "<tr><td>&nbsp;</td></tr>";
  
  txt += "<tr valign=top><td>color</td>";
  txt += "<td>";
  txt += "<input type=radio name=color value = red ";     if(color == "red") txt += "checked"; txt += ">red<BR>";
  txt += "<input type=radio name=color value = orange ";  if(color == "orange") txt += "checked"; txt += ">orange<BR>";
  txt += "<input type=radio name=color value = brown ";   if(color == "brown") txt += "checked"; txt += ">brown<BR>";
  txt += "<input type=radio name=color value = blue ";    if(color == "blue") txt += "checked"; txt += ">blue<BR>";
  txt += "<input type=radio name=color value = green ";   if(color == "green") txt += "checked"; txt += ">green<BR>";
  txt += "<input type=radio name=color value = purple ";  if(color == "purple") txt += "checked"; txt += ">purple<BR>";
  txt += "<input type=radio name=color value = white ";   if(color == "white") txt += "checked"; txt += ">white<BR>";
  txt += "</td></tr>";
  txt += "<tr><td>&nbsp;</td></tr>";
 
  txt += "<tr valign=top><td>DST (summer / winter time)&nbsp;</td>";
  txt += "<td>";
  txt += "<input type=radio name=DST_mode value = summer "; if(DST_mode == "summer")  txt += "checked"; txt += ">summer<BR>";
  txt += "<input type=radio name=DST_mode value = winter "; if(DST_mode == "winter")  txt += "checked"; txt += ">winter<BR>";
  txt += "<input type=radio name=DST_mode value = auto ";   if(DST_mode == "auto")    txt += "checked"; txt += ">auto "; if (summer) txt += "(summer)"; else txt += "(winter)";
  txt += "</td></tr>";
  txt += "<tr><td>&nbsp;</td></tr>";
  
  txt += "<tr valign=top><td>NTP offset summer (EU = 1)</td>";
  txt += "<td>";
  txt += "<input type=text name=summer_offset value = " + (String) summer_offset + ">";
  txt += "</td></tr>";
  txt += "<tr valign=top><td>NTP offset winter (EU = 0)</td>";
  txt += "<td>";
  txt += "<input type=text name=winter_offset value = " + (String) winter_offset + ">";
  txt += "</td></tr>";
  txt += "<tr><td>&nbsp;</td></tr>";

  txt += "<tr valign=top><td>Accesspoint name</td>";
  txt += "<td><input type=text name=Accesspoint_Name maxlength=19 value = " + Accesspoint_Name + ">";
  txt += "</td></tr>";

  txt += "<tr valign=top><td>Accesspoint ip address</td>";
  txt += "<td>192.168.4.1";
  txt += "</td></tr>";
  
  txt += "</table>";
  
  txt += "<br><br>";
  txt += "<input type=submit name=ok value=ok>&nbsp;&nbsp;<input type=submit name=cancel value=cancel>";
  txt += "</form></font></body></html>";

  return txt;
}
